<?php
namespace Tagalys\Sync\Helper;

class Category extends \Magento\Framework\App\Helper\AbstractHelper
{
    private $updatedCategories = array();
    public function __construct(
        \Tagalys\Sync\Helper\Configuration $tagalysConfiguration,
        \Tagalys\Sync\Helper\Api $tagalysApi,
        \Magento\Catalog\Model\ResourceModel\Category\Collection $categoryCollection,
        \Magento\Catalog\Model\ResourceModel\Category\CollectionFactory $categoryCollectionFactory,
        \Tagalys\Sync\Model\CategoryFactory $tagalysCategoryFactory,
        \Magento\Catalog\Model\CategoryFactory $categoryFactory,
        \Magento\Store\Model\StoreManagerInterface $storeManagerInterface,
        \Magento\Catalog\Model\ProductFactory $productFactory,
        \Magento\Framework\App\ResourceConnection $resourceConnection,
        \Magento\Framework\Model\ResourceModel\Iterator $resourceModelIterator,
        \Magento\Catalog\Api\Data\CategoryProductLinkInterfaceFactory $categoryProductLinkInterfaceFactory,
        \Magento\Catalog\Api\CategoryLinkRepositoryInterface $categoryLinkRepositoryInterface,
        \Magento\Framework\App\CacheInterface $cacheInterface,
        \Magento\Framework\Event\Manager $eventManager,
        \Magento\Framework\App\ProductMetadataInterface $productMetadataInterface,
        \Magento\Indexer\Model\IndexerFactory $indexerFactory,
        \Magento\Framework\Math\Random $random,
        \Magento\Framework\Registry $registry,
        \Magento\UrlRewrite\Model\UrlRewriteFactory $urlRewriteFactory,
        \Magento\UrlRewrite\Model\ResourceModel\UrlRewriteCollection $urlRewriteCollection
    ) {
        $this->tagalysConfiguration = $tagalysConfiguration;
        $this->random = $random;
        $this->_registry = $registry;
        $this->tagalysApi = $tagalysApi;
        $this->categoryCollection = $categoryCollection;
        $this->categoryCollectionFactory = $categoryCollectionFactory;
        $this->tagalysCategoryFactory = $tagalysCategoryFactory;
        $this->productFactory = $productFactory;
        $this->categoryFactory = $categoryFactory;
        $this->storeManagerInterface = $storeManagerInterface;
        $this->resourceConnection = $resourceConnection;
        $this->resourceModelIterator = $resourceModelIterator;
        $this->categoryProductLinkInterfaceFactory = $categoryProductLinkInterfaceFactory;
        $this->categoryLinkRepositoryInterface = $categoryLinkRepositoryInterface;
        $this->cacheInterface = $cacheInterface;
        $this->eventManager = $eventManager;
        $this->productMetadataInterface = $productMetadataInterface;
        $this->indexerFactory = $indexerFactory;
        $this->urlRewriteFactory = $urlRewriteFactory;
        $this->urlRewriteCollection = $urlRewriteCollection;
        
        $writer = new \Zend\Log\Writer\Stream(BP . '/var/log/tagalys_commands.log');
        $this->tagalysCommandsLogger = new \Zend\Log\Logger();
        $this->tagalysCommandsLogger->addWriter($writer);
        $writer = new \Zend\Log\Writer\Stream(BP . '/var/log/tagalys_categories.log');
        $this->logger = new \Zend\Log\Logger();
        $this->logger->addWriter($writer);
    }

    public function truncate() {
        $tagalysCategoryFactory = $this->tagalysCategoryFactory->create();
        $connection = $tagalysCategoryFactory->getResource()->getConnection();
        $tableName = $tagalysCategoryFactory->getResource()->getMainTable();
        $connection->truncateTable($tableName);
    }

    public function createOrUpdateWithData($storeId, $categoryId, $createData, $updateData)
    {
        $firstItem = $this->tagalysCategoryFactory->create()->getCollection()
        ->addFieldToFilter('category_id', $categoryId)
        ->addFieldToFilter('store_id', $storeId)
        ->getFirstItem();

        try {
        if ($id = $firstItem->getId()) {
            $updateData['category_id'] = $categoryId;
            $updateData['store_id'] = $storeId;
            $model = $this->tagalysCategoryFactory->create()->load($id)->addData($updateData);
            $model->setId($id)->save();
        } else {
            $createData['category_id'] = $categoryId;
            $createData['store_id'] = $storeId;
            $model = $this->tagalysCategoryFactory->create()->setData($createData);
            $insertId = $model->save()->getId();
        }
        } catch (Exception $e) {
        
        }
    }
    public function markStoreCategoryIdsForDeletionExcept($storeId, $categoryIds) {
        $collection = $this->tagalysCategoryFactory->create()->getCollection()->addFieldToFilter('store_id', $storeId);
        foreach ($collection as $collectionItem) {
        if (!in_array((int)$collectionItem->getCategoryId(), $categoryIds)) {
            $collectionItem->addData(array('marked_for_deletion' => 1))->save();
        }
        }
    }

    public function isMultiStoreWarningRequired()
    {
        $allStores = $this->tagalysConfiguration->getAllWebsiteStores();
        $showMultiStoreConfig = false;
        $checkAllCategories = false;
        if (count($allStores) > 1) {
        $rootCategories = array();
        foreach ($allStores as $store) {
            $rootCategoryId = $this->storeManagerInterface->getStore($store['value'])->getRootCategoryId();
            if (in_array($rootCategoryId, $rootCategories)) {
            $checkAllCategories = true;
            break;
            } else {
            array_push($rootCategories, $rootCategoryId);
            }
        }
        }

        if ($checkAllCategories) {
        $allCategories = array();
        foreach ($allStores as $store) {
            $rootCategoryId = $this->storeManagerInterface->getStore($store['value'])->getRootCategoryId();
            $categories = $this->categoryCollection
            ->setStoreId($store['value'])
            ->addFieldToFilter('is_active', 1)
            ->addAttributeToFilter('path', array('like' => "1/{$rootCategoryId}/%"))
            ->addAttributeToSelect('id');
            foreach ($categories as $cat) {
            if (in_array($cat->getId(), $allCategories)) {
                $showMultiStoreConfig = true;
                break;
            } else {
                array_push($allCategories, $cat->getId());
            }
            }
        }
        }
        return $showMultiStoreConfig;
    }

    public function transitionFromCategoriesConfig()
    {
        $categoryIds = $this->tagalysConfiguration->getConfig("category_ids", true);
        $storesForTagalys = $this->tagalysConfiguration->getStoresForTagalys();
        foreach ($storesForTagalys as $storeId) {
        $originalStoreId = $this->storeManagerInterface->getStore();
        $this->storeManagerInterface->setCurrentStore($storeId);
        foreach ($categoryIds as $i => $categoryId) {
            $category = $this->categoryFactory->create()->load($categoryId);
            $categoryActive = $category->getIsActive();
            if ($categoryActive && ($category->getDisplayMode() != 'PAGE')) {
            // TODO: createOrUpdateWithData() & Mage_Catalog_Model_Category::DM_PAGE -> PAGE
            $this->createOrUpdateWithData($storeId, $categoryId, array('positions_sync_required' => 0, 'marked_for_deletion' => 0, 'status' => 'pending_sync'), array('marked_for_deletion' => 0));
            }
        }
        $this->storeManagerInterface->setCurrentStore($originalStoreId);
        }
    }

    public function isProductPushDownAllowed($categoryId)
    {
        $allStores = $this->storeManagerInterface->getStores();
        $activeInStores = 0;
        if (count($allStores) == 1) {
        // Single Store
        return true;
        }
        // Multiple Stores
        foreach ($allStores as $store) {
        $categories = $this->categoryCollection
            ->setStoreId($store['value'])
            ->addFieldToFilter('is_active', 1)
            ->addFieldToFilter('entity_id', $categoryId)
            ->addAttributeToSelect('id');
        if ($categories->count() > 0) {
            $activeInStores++;
            if ($activeInStores > 1) {
            return ($this->tagalysConfiguration->getConfig("listing_pages:same_or_similar_products_across_all_stores") == '1');
            }
        }
        }
        return true;
    }

    public function maintenanceSync()
    {
        // once a day
        $listingPagesEnabled = ($this->tagalysConfiguration->getConfig("module:listingpages:enabled") == '1');
        if ($listingPagesEnabled) {
        // 1. try and sync all failed categories - mark positions_sync_required as 1 for all failed categories - this will then try and sync the categories again
        $failedCategories = $this->tagalysCategoryFactory->create()->getCollection()
            ->addFieldToFilter('status', 'failed')
            ->addFieldToFilter('marked_for_deletion', 0);
        foreach ($failedCategories as $i => $failedCategory) {
            $failedCategory->addData(array('status' => 'pending_sync'))->save();
        }

        // 2. if preference is to power all categories, loop through all categories and add missing items to the tagalys_core_categories table
        // TODO
        // 3. send all category ids to be powered by tagalys - tagalys will delete other ids
        $storesForTagalys = $this->tagalysConfiguration->getStoresForTagalys();
        $categoriesForTagalys = array();
        foreach ($storesForTagalys as $key => $storeId) {
            $categoriesForTagalys[$storeId] = array();
            $storeCategories = $this->tagalysCategoryFactory->create()->getCollection()
            ->addFieldToFilter('store_id', $storeId);
            foreach ($storeCategories as $i => $storeCategory) {
            array_push($categoriesForTagalys[$storeId], '__categories--' . $storeCategory->getCategoryId());
            }
        }
        $this->tagalysApi->clientApiCall('/v1/mpages/_platform/verify_enabled_pages', array('enabled_pages' => $categoriesForTagalys));
        }
        return true;
    }

    public function _checkSyncLock($syncStatus)
    {
        if ($syncStatus['locked_by'] == null) {
        return true;
        } else {
        // some other process has claimed the thread. if a crash occours, check last updated at < 15 minutes ago and try again.
        $lockedAt = new \DateTime($syncStatus['updated_at']);
        $now = new \DateTime();
        $intervalSeconds = $now->getTimestamp() - $lockedAt->getTimestamp();
        $minSecondsForOverride = 5 * 60;
        if ($intervalSeconds > $minSecondsForOverride) {
            $this->tagalysApi->log('warn', 'Overriding stale locked process for categories sync', array('pid' => $syncStatus['locked_by'], 'locked_seconds_ago' => $intervalSeconds));
            return true;
        } else {
            $this->tagalysApi->log('warn', 'Categories sync locked by another process', array('pid' => $syncStatus['locked_by'], 'locked_seconds_ago' => $intervalSeconds));
            return false;
        }
        }
    }

    public function markPositionsSyncRequired($storeId, $categoryId)
    {
        $firstItem = $this->tagalysCategoryFactory->create()->getCollection()
        ->addFieldToFilter('store_id', $storeId)
        ->addFieldToFilter('category_id', $categoryId)
        ->getFirstItem();
        if ($id = $firstItem->getId()) {
        $firstItem->addData(array('positions_sync_required' => 1))->save();
        }
        return true;
    }
    public function markPositionsSyncRequiredForCategories($storeId, $categoryIds) {
        $conn = $this->resourceConnection->getConnection();
        $tableName = $this->resourceConnection->getTableName('tagalys_category');
        $whereData = array(
            'status = ?' => 'powered_by_tagalys'
        );
        if ($storeId !== 'all') {
            $whereData['store_id = ?'] = $storeId;
        }
        if ($categoryIds !== 'all') {
            $whereData['category_id IN (?)'] = $categoryIds;
        }
        $updateData = array(
            'positions_sync_required' => 1
        );
        $conn->update($tableName, $updateData, $whereData);
        return true;
    }
    public function markFailedCategoriesForRetrying() {
        $conn = $this->resourceConnection->getConnection();
        $tableName = $this->resourceConnection->getTableName('tagalys_category');
        $whereData = array(
        'status = ?' => 'failed'
        );
        $updateData = array(
        'status' => 'pending_sync'
        );
        $conn->update($tableName, $updateData, $whereData);
        return true;
    }
    public function getEnabledCount($storeId)
    {
        return $this->tagalysCategoryFactory->create()->getCollection()
        ->addFieldToFilter('store_id', $storeId)
        ->addFieldToFilter('marked_for_deletion', 0)
        ->count();
    }
    public function getPendingSyncCount($storeId)
    {
        return $this->tagalysCategoryFactory->create()->getCollection()
        ->addFieldToFilter('store_id', $storeId)
        ->addFieldToFilter('status', 'pending_sync')
        ->addFieldToFilter('marked_for_deletion', 0)
        ->count();
    }
    public function getRequiringPositionsSyncCount($storeId)
    {
        return $this->tagalysCategoryFactory->create()->getCollection()
        ->addFieldToFilter('store_id', $storeId)
        ->addFieldToFilter('status', 'powered_by_tagalys')
        ->addFieldToFilter('positions_sync_required', 1)
        ->addFieldToFilter('marked_for_deletion', 0)
        ->count();
    }
    public function getRequiresPositionsSyncCollection()
    {
        $categoriesToSync = $this->tagalysCategoryFactory->create()->getCollection()
        ->addFieldToFilter('status', 'powered_by_tagalys')
        ->addFieldToFilter('positions_sync_required', 1)
        ->addFieldToFilter('marked_for_deletion', 0);
        return $categoriesToSync;
    }

    public function updatePositionsIfRequired($maxProductsPerCronRun = 50, $perPage = 5, $force = false)
    {
        $listingPagesEnabled = ($this->tagalysConfiguration->getConfig("module:listingpages:enabled") == '1');
        if ($listingPagesEnabled || $force) {
            $pid = $this->random->getRandomString(24);
            $this->tagalysApi->log('local', '1. Started updatePositionsIfRequired', array('pid' => $pid));
            $categoriesSyncStatus = $this->tagalysConfiguration->getConfig("categories_sync_status", true);
            if ($this->_checkSyncLock($categoriesSyncStatus)) {
                $utcNow = new \DateTime("now", new \DateTimeZone('UTC'));
                $timeNow = $utcNow->format(\DateTime::ATOM);
                $syncStatus = array(
                'updated_at' => $timeNow,
                'locked_by' => $pid
                );
                $this->tagalysConfiguration->setConfig('categories_sync_status', $syncStatus, true);
                $collection = $this->getRequiresPositionsSyncCollection();
                $remainingCount = $collection->count();
                // $this->logger("updatePositionsIfRequired: remainingCount: {$remainingCount}", null, 'tagalys_processes.log', true);
                $countToSyncInCronRun = min($remainingCount, $maxProductsPerCronRun);
                $numberCompleted = 0;
                $circuitBreaker = 0;
                while ($numberCompleted < $countToSyncInCronRun && $circuitBreaker < 26) {
                $circuitBreaker += 1;
                $categoriesToSync = $this->getRequiresPositionsSyncCollection()->setPageSize($perPage);
                $utcNow = new \DateTime("now", new \DateTimeZone('UTC'));
                $timeNow = $utcNow->format(\DateTime::ATOM);
                $syncStatus['updated_at'] = $timeNow;
                $this->tagalysConfiguration->setConfig('categories_sync_status', $syncStatus, true);
                foreach ($categoriesToSync as $categoryToSync) {
                    $storeId = $categoryToSync->getStoreId();
                    $categoryId = $categoryToSync->getCategoryId();
                    $newPositions = $this->tagalysApi->storeApiCall($storeId . '', '/v1/mpages/_platform/__categories-' . $categoryId . '/positions', array());
                    if ($newPositions != false) {
                    if($this->tagalysConfiguration->getConfig('listing_pages:position_sort_direction') == 'asc'){
                        $this->_updatePositions($categoryId, $newPositions['positions']);
                    } else {
                        $this->_updatePositionsReverse($categoryId, $newPositions['positions']);
                    }
                    array_push($this->updatedCategories, $categoryId);
                    $categoryToSync->addData(array('positions_sync_required' => 0, 'positions_synced_at' => date("Y-m-d H:i:s")))->save();
                    } else {
                    // api call failed
                    }
                }
                $numberCompleted += $categoriesToSync->count();
                // $this->logger("updatePositionsIfRequired: completed {$numberCompleted}", null, 'tagalys_processes.log', true);
                }
                $syncStatus['locked_by'] = null;
                $this->tagalysConfiguration->setConfig('categories_sync_status', $syncStatus, true);

                $this->processUpdatedCategories(true);
            }
        }
    }

    public function _updatePositions($categoryId, $positions)
    {
        if ($this->isProductPushDownAllowed($categoryId)) {
            $whereData = array(
                'category_id = ?' => (int)$categoryId,
                'position <= ?' => count($positions)
            );
            $updateData = array(
                'position' => (count($positions) + 1)
            );
            $this->runSqlForCategoryPositions($updateData, $whereData);
            }

            foreach ($positions as $productId => $productPosition) {
            $whereData = array(
                'category_id = ?' => (int)$categoryId,
                'product_id = ?' => (int)$productId
            );
            $updateData = array(
                'position' => (int)$productPosition
            );
            $this->runSqlForCategoryPositions($updateData, $whereData);
        }
        return true;
    }

    public function _updatePositionsReverse($categoryId, $positions)
    {
        if ($this->isProductPushDownAllowed($categoryId)) {
            $whereData = array(
                'category_id = ?' => (int)$categoryId,
                'position >= ?' => 100
            );
            $updateData = array(
                'position' => 99
            );
            $this->runSqlForCategoryPositions($updateData, $whereData);
            }

            $totalPositions = count($positions);
            foreach ($positions as $productId => $productPosition) {
            $whereData = array(
                'category_id = ?' => (int)$categoryId,
                'product_id = ?' => (int)$productId
            );
            $updateData = array(
                'position' => 101 + $totalPositions - (int)$productPosition
            );
            $this->runSqlForCategoryPositions($updateData, $whereData);
        }
        return true;
    }

    public function getRemainingForSync()
    {
        return $this->tagalysCategoryFactory->create()->getCollection()
        ->addFieldToFilter('status', 'pending_sync')
        ->addFieldToFilter('marked_for_deletion', 0)->count();
    }
    public function getRemainingForDelete()
    {
        return $this->tagalysCategoryFactory->create()->getCollection()
        ->addFieldToFilter('marked_for_deletion', 1)->count();
    }
    public function syncAll($force = false)
    {
        $remainingForSync = $this->getRemainingForSync();
        $remainingForDelete = $this->getRemainingForDelete();
        // echo('syncAll: ' . json_encode(compact('remainingForSync', 'remainingForDelete')));
        while ($remainingForSync > 0 || $remainingForDelete > 0) {
        $this->sync(50, $force);
        $remainingForSync = $this->getRemainingForSync();
        $remainingForDelete = $this->getRemainingForDelete();
        // echo('syncAll: ' . json_encode(compact('remainingForSync', 'remainingForDelete')));
        }
    }
    public function getStoreCategoryDetails($storeId, $categoryId) {
        $originalStoreId = $this->storeManagerInterface->getStore()->getId();
        $this->storeManagerInterface->setCurrentStore($storeId);
        $category = null;
        $category = $this->categoryFactory->create()->load($categoryId);
        $categoryActive = $category->getIsActive();
        if ($categoryActive) {
            return array(
                "id" => "__categories-$categoryId",
                "slug" => $category->getUrl(),
                "enabled" => true,
                "name" => implode(' / ', array_slice(explode(' |>| ', $this->tagalysConfiguration->getCategoryName($category)), 1)),
                "filters" => array(
                array(
                    "field" => "__categories",
                    "value" => $categoryId
                ),
                array(
                    "field" => "visibility",
                    "tag_jsons" => array("{\"id\":\"2\",\"name\":\"Catalog\"}", "{\"id\":\"4\",\"name\":\"Catalog, Search\"}")
                )
            ));
        }
        $this->storeManagerInterface->setCurrentStore($originalStoreId);
    }
    public function sync($max, $force = false)
    {
        $listingPagesEnabled = ($this->tagalysConfiguration->getConfig("module:listingpages:enabled") == '1');
        if ($listingPagesEnabled || $force) {
        $detailsToSync = array();

        // save
        $categoriesToSync = $this->tagalysCategoryFactory->create()->getCollection()
            ->addFieldToFilter('status', 'pending_sync')
            ->addFieldToFilter('marked_for_deletion', 0)
            ->setPageSize($max);
        foreach ($categoriesToSync as $i => $categoryToSync) {
            $storeId = $categoryToSync->getStoreId();
            array_push($detailsToSync, array('perform' => 'save', 'store_id' => $storeId, 'payload' => $this->getStoreCategoryDetails($storeId, $categoryToSync->getCategoryId())));
        }
        // delete
        $categoriesToDelete = $this->tagalysCategoryFactory->create()->getCollection()
            ->addFieldToFilter('marked_for_deletion', 1)
            ->setPageSize($max);
        foreach ($categoriesToDelete as $i => $categoryToDelete) {
            $storeId = $categoryToDelete->getStoreId();
            $categoryId = $categoryToDelete->getCategoryId();
            array_push($detailsToSync, array('perform' => 'delete', 'store_id' => $storeId, 'payload' => array('id' => "__categories-{$categoryId}")));
        }

        if (count($detailsToSync) > 0) {
            // sync
            $tagalysResponse = $this->tagalysApi->clientApiCall('/v1/mpages/_sync_platform_pages', array('actions' => $detailsToSync));

            if ($tagalysResponse != false) {
            foreach ($tagalysResponse['save_actions'] as $i => $saveActionResponse) {
                $firstItem = $this->tagalysCategoryFactory->create()->getCollection()
                ->addFieldToFilter('store_id', $saveActionResponse['store_id'])
                ->addFieldToFilter('category_id', explode('-', $saveActionResponse['id'])[1])
                ->getFirstItem();
                if ($id = $firstItem->getId()) {
                if ($saveActionResponse['saved']) {
                    $firstItem->addData(array('status' => 'powered_by_tagalys', 'positions_sync_required' => 1))->save();
                } else {
                    $firstItem->addData(array('status' => 'failed'))->save();
                }
                }
            }
            foreach ($categoriesToDelete as $i => $categoryToDelete) {
                $categoryToDelete->delete();
            }
            }
        }
        }
    }

    public function assignParentCategoriesToAllProducts($viaDb = false){
        $productCollection = $this->productFactory->create()->getCollection()
                        ->addAttributeToFilter('status', 1)
                        ->addAttributeToFilter('visibility', array("neq" => 1))
                        ->addAttributeToSelect('entity_id, product_id')
                        ->load();
        $this->resourceModelIterator->walk($productCollection->getSelect(), array(array($this, 'assignParentCategoriesToProductHandler')), array('viaDb' => $viaDb));
    }

    public function assignParentCategoriesToProductHandler($args){
        $this->assignParentCategoriesToProductId($args['row']['entity_id'], $args['viaDb']);
    }

    public function assignParentCategoriesToProductId($productId, $viaDb = false) {
        $this->tagalysCommandsLogger->info( "assignParentCategoriesToProductId: $productId");
        $product = $this->productFactory->create()->load($productId);
        $categoryIds = $product->getCategoryIds();
        $assignedParents = array();
        foreach($categoryIds as $categoryId){
            $category = $this->categoryFactory->create()->load($categoryId);
            foreach($category->getParentCategories() as $parent){
                if ((int)$parent->getLevel() > 1) {
                    if(!in_array($parent->getId(), $categoryIds) and !in_array($parent->getId(), $assignedParents)){
                        array_push($assignedParents, $parent->getId());
                        if ($viaDb) {
                            $this->assignProductToCategoryViaDb($parent->getId(), $product);
                        } else {
                            $this->assignProductToCategory($parent->getId(), $product);
                        }
                    }
                }
            }
        }
    }

    public function assignProductToCategory($categoryId, $product, $viaDb = false){
        $this->tagalysCommandsLogger->info("assignProductToCategory: {$categoryId}");
        $productSku = $product->getSku();
        $categoryProductLink = $this->categoryProductLinkInterfaceFactory->create();
        $categoryProductLink->setSku($productSku);
        $categoryProductLink->setCategoryId($categoryId);
        $categoryProductLink->setPosition(999);
        $this->categoryLinkRepositoryInterface->save($categoryProductLink);
    }
    public function assignProductToCategoryViaDb($categoryId, $product){
        $this->tagalysCommandsLogger->info("assignProductToCategoryViaDb: {$categoryId}");
        $conn = $this->resourceConnection->getConnection();
        $table = $this->resourceConnection->getTableName('catalog_category_product');
        
        $sortDirection = $this->tagalysConfiguration->getConfig('listing_pages:position_sort_direction');
        $positionToAssign = ($sortDirection == 'desc' ? 1 : 9999);
        $assignData = array('category_id'=>(int)$categoryId, 'product_id'=>(int)($product->getId()), 'position' => $positionToAssign);
        $conn->insert($table, $assignData);
    }

    public function uiPoweredByTagalys($storeId, $categoryId) {
        try {
            $firstItem = $this->tagalysCategoryFactory->create()->getCollection()
                ->addFieldToFilter('store_id', $storeId)
                ->addFieldToFilter('category_id', $categoryId)
                ->addFieldToFilter('status', 'powered_by_tagalys')
                ->addFieldToFilter('marked_for_deletion', 0)
                ->getFirstItem();
            if ($id = $firstItem->getId()) {
                return true;
            } else {
                return false;
            }
        } catch (Exception $e) {
            return false;
        }
    }
        public function pushDownProductsIfRequired($productIds, $productCategories = null, $indexerToRun = 'product') {
        // called from observers when new products are added to categories - in position ascending order, they should be positioned at the bottom of the page.
        $listingpagesEnabled = $this->tagalysConfiguration->getConfig('module:listingpages:enabled');
        if($listingpagesEnabled == '1') {
            $sortDirection = $this->tagalysConfiguration->getConfig('listing_pages:position_sort_direction');
            if($sortDirection == 'asc'){
                $tagalysCategories = $this->getTagalysCategories($productCategories);
                if (count($tagalysCategories) > 0) {
                    $whereData = array(
                        'category_id IN (?)' => $tagalysCategories,
                        'product_id IN (?)' => $productIds,
                        'position = ?' => 0
                    );
                    $updateData = array(
                        'position' => 9999
                    );
                    $this->runSqlForCategoryPositions($updateData, $whereData);
                    $categoryProductIndexer = $this->indexerFactory->create()->load('catalog_category_product');
                    if ($categoryProductIndexer->isScheduled() === false) {
                    // reindex if index mode is update on save. not required for schedule because we don't want these showing up so it has the same effect. when tagalys syncs and updates positions, reindex will be triggered.
                    switch($indexerToRun) {
                        case 'product':
                        $reindexAndClearCacheImmediately = $this->tagalysConfiguration->getConfig('listing_pages:reindex_and_clear_cache_immediately');
                        if ($reindexAndClearCacheImmediately == '1') {
                            $indexer = $this->indexerFactory->create()->load('catalog_product_category');
                            $indexer->reindexList($productIds);
                        }
                        // cache may already be getting updated. even otherwise, we don't need to update here as we don't want products to show up anyway. caches are cleared when positions are updated via Tagalys.
                        break;
                        case 'category':
                        array_push($this->updatedCategories, $tagalysCategories);
                        $this->processUpdatedCategories();
                        break;
                    }
                    }
                }
            }
        }
    }

    public function processUpdatedCategories($updateSortOrder = false) {
        $writer = new \Zend\Log\Writer\Stream(BP . '/var/log/tagalys_commands.log');
        $this->tagalysCommandsLogger = new \Zend\Log\Logger();
        $this->tagalysCommandsLogger->addWriter($writer);
        if (count($this->updatedCategories) > 0) {

            if ($updateSortOrder) {
                // set sort order to position
                foreach ($this->updatedCategories as $categoryId) {
                // don't set default store order because that could change all store's sort orders
                    // $category = $this->categoryFactory->create()->setStoreId('0')->load($categoryId);
                    // $category->setDefaultSortBy('position')->save();

                // set position sort order for all tagalys stores for all categories. assumption: if a category is assigned to tagalys in one store, it has to be assigned to tagalys in other stores that are powered by tagalys.
                $storesForTagalys = $this->tagalysConfiguration->getStoresForTagalys();
                foreach ($storesForTagalys as $storeId) {
                    $category = $this->categoryFactory->create()->setStoreId($storeId)->load($categoryId);
                    if ($category->getDefaultSortBy() != 'position') {
                        $this->setDefaultSortBy($categoryId);
                    }
                }
                }
            }

            $reindexAndClearCacheImmediately = $this->tagalysConfiguration->getConfig('listing_pages:reindex_and_clear_cache_immediately');
            if ($reindexAndClearCacheImmediately == '1') {
                // reindex
                $indexer = $this->indexerFactory->create()->load('catalog_category_product');
                $indexer->reindexList($this->updatedCategories);

                foreach($this->updatedCategories as $categoryId) {
                // clear magento cache
                $category = $this->categoryFactory->create()->load($categoryId);
                $this->eventManager->dispatch('clean_cache_by_tags', ['object' => $category]);
                }
            }

            // trigger event to allow clearing custom cache
            $categoryIdsObj = new \Magento\Framework\DataObject(array('category_ids' => $this->updatedCategories));
            $this->eventManager->dispatch('tagalys_category_positions_updated', ['tgls_data' => $categoryIdsObj]);

            $this->updatedCategories = array();
        }
    }

    public function runSqlForCategoryPositions($updateData, $whereData) {
        $conn = $this->resourceConnection->getConnection();
        $tableName = $this->resourceConnection->getTableName('catalog_category_product');
        $conn->update($tableName, $updateData, $whereData);
    }

    public function getTagalysCategories($categoryIds = null) {
        $conn = $this->resourceConnection->getConnection();
        $tableName = $this->resourceConnection->getTableName('tagalys_category');
        $select = $conn->select()->from($tableName)->where('marked_for_deletion = ? and status != "failed"', 0);
        if($categoryIds!=null && is_array($categoryIds) && count($categoryIds) > 0 ){
        $select->where('category_id IN (?)', $categoryIds);
        }
        $result = $conn->fetchAll($select);
        $tagalysCategories = array();
        foreach($result as $row){
            $tagalysCategories[] = $row['category_id'];
        }
        $tagalysCategories = array_unique($tagalysCategories);
        return $tagalysCategories;
    }
  
    public function performCategoryPositionUpdate($categoryId, $positions) {
        $sortOrder = $this->tagalysConfiguration->getConfig('listing_pages:position_sort_direction');
        if($sortOrder == 'asc') {
            $this->_updatePositions($categoryId, $positions);
        } else {
            $this->_updatePositionsReverse($categoryId, $positions);
        }
        $this->reindexUpdatedCategories($categoryId);
        return true;
    }
    
    public function createTagalysParentCategory($storeId, $categoryDetails) {
        $rootCategoryId = $this->storeManagerInterface->getStore($storeId)->getRootCategoryId();
        $categoryDetails['is_active'] = false;
        $categoryId = $this->_createCategory($rootCategoryId, $categoryDetails);
        if($categoryId){
            $this->tagalysConfiguration->setConfig("tagalys_parent_category_store_$storeId", $categoryId);
            return $categoryId;
        }
        return false;
    }

    public function createCategory($storeId, $categoryDetails) {
        $parentCategoryId = $this->tagalysConfiguration->getConfig("tagalys_parent_category_store_$storeId");
        $parentCategoryId = $this->categoryFactory->create()->load($parentCategoryId)->getId();
        if ($parentCategoryId == null) {
            throw new \Exception("Tagalys parent category not created. Please enable Smart Categories in Tagalys Configuration.");
        }
        $categoryDetails['is_active'] = true;
        $categoryDetails['include_in_menu'] = false;
        $categoryDetails['default_sort_by'] = 'position';
        $categoryDetails['display_mode'] = \Magento\Catalog\Model\Category::DM_PRODUCT;
        $categoryId = $this->_createCategory($parentCategoryId, $categoryDetails);
        return $categoryId;
    }

    public function updateCategoryDetails($categoryId, $categoryDetails) {
        $this->logger->info("updateCategoryDetails: category_id: $categoryId, categoryDetails: ".json_encode($categoryDetails));
        $category = $this->categoryFactory->create()->load($categoryId);
        if($category->getId() == null){
            throw new \Exception("Platform category not found");
        }
        $categoryDetails['default_sort_by'] = 'position';
        $category->setStoreId(0)->addData($categoryDetails);
        $category->save();
        $this->categoryUpdateAfter($category);
        return true;
    }

    public function categoryUpdateAfter($category){
        // TODO: Remove in future releases
        if($this->isLegacyMpageCategory($category)){
            $this->logger->info("categoryUpdateAfter: Updating url_rewrite for legacy mpage category, category_id: ".$category->getId());
            $urlRewrite = $this->urlRewriteCollection->addFieldToFilter('target_path', "catalog/category/view/id/{$category->getId()}")->getFirstItem();
            $urlRewrite->setRequestPath("m/{$category->getUrlKey()}")->save();
        }
    }

    public function deleteTagalysCategory($categoryId) {
        $category = $this->categoryFactory->create()->load($categoryId);
        if($category->getId() === null){
            return true;
        }
        $allowDelete = $this->isTagalysCreated($categoryId);
        if($allowDelete){
            $this->_registry->register("isSecureArea", true);
            $category->delete();
            return true;
        }
        throw new \Exception("This category cannot be deleted because it wasn't created by Tagalys");
    }

    public function _createCategory($parentId, $categoryDetails){
        $this->logger->info("_createCategory: parent_id: $parentId, categoryDetails: ".json_encode($categoryDetails));
        $category = $this->categoryFactory->create();
        $parent = $this->categoryFactory->create()->load($parentId);
        $category->setData($categoryDetails);
        $category->addData([
            'parent_id' => $parentId,
            'path' => $parent->getPath(),
            'default_sort_by' => 'position',
            'display_mode' => \Magento\Catalog\Model\Category::DM_PRODUCT,
            'include_in_menu' => 0,
            'is_anchor' => 1
        ]);
        $category->setStoreId(0);
        $category->save();
        $categoryId = $this->categoryCollectionFactory->create()
            ->addAttributeToFilter('parent_id', $parentId)
            ->addAttributeToFilter('url_key', $categoryDetails['url_key'])
            ->addAttributeToSelect('entity_id')
            ->getFirstItem()
            ->getId();
        return $categoryId;
    }

    public function setDefaultSortBy($categoryId)
    {
        $this->logger->info("setDefaultSortBy: category_id: $categoryId");
        $category = $this->categoryFactory->create()->load($categoryId);
        if ($category->getId() == null) {
            throw new \Exception("Platform category not found");
        }
        $entityTypeTable = $this->resourceConnection->getTableName('eav_entity_type');
        $eavAttribute = $this->resourceConnection->getTableName('eav_attribute');
        $cev = $this->resourceConnection->getTableName('catalog_category_entity_varchar');
        $conn = $this->resourceConnection->getConnection();
        $select = $conn->fetchAll("SELECT entity_type_id FROM $entityTypeTable WHERE entity_type_code='catalog_category';");
        $entityTypeId = $select[0]['entity_type_id'];
        $select = $conn->fetchAll("SELECT attribute_code, attribute_id FROM $eavAttribute WHERE entity_type_id=$entityTypeId AND attribute_code IN ('default_sort_by');");
        $defaultSortByAttribute = $select[0]['attribute_id'];
        $updateData = ['value' => 'position'];
        $where = ['entity_id = ?' => $categoryId, 'attribute_id = ?' => $defaultSortByAttribute];
        $conn->update($cev, $updateData, $where);

        $stores = $this->storeManagerInterface->getStores();
        foreach ($stores as $storeId => $store) {
            try{
                $flatTable = $this->resourceConnection->getTableName("catalog_category_flat_store_$storeId");
                $updateData = ['default_sort_by' => 'position'];
                $where = ['entity_id = ?' => $categoryId];
                $conn->update($flatTable, $updateData, $where);
            } catch(\Exception $e){
                // Ignoring base table not found exception
            }
        }
        return true;
    }

    public function bulkAssignProductsToCategoryAndRemove($categoryId, $productPositions) {
        if($this->isTagalysCreated($categoryId)){
          $productsToRemove = $this->getProductsNotIn($categoryId, $productPositions);
          $this->bulkAssignProductsToCategoryViaDb($categoryId, $productPositions);
          $this->_paginateSqlRemove($categoryId, $productsToRemove);
          $this->reindexUpdatedCategories();
          return true;
        }
        throw new \Exception("Error: this category wasn't created by Tagalys");
    }

    private function getProductsNotIn($categoryId, $productPositions){
        $tableName = $this->resourceConnection->getTableName('catalog_category_product');
        $sql = "SELECT product_id FROM $tableName WHERE category_id=$categoryId";
        $select = $this->runSqlSelect($sql);
        $productsInTable = [];
        foreach($select as $row){
            $productsInTable[] = $row['product_id'];
        }
        return array_diff($productsInTable, $productPositions);
    }

    public function bulkAssignProductsToCategoryViaDb($categoryId, $productPositions) {
        if(count($productPositions)>0){
            if($this->tagalysConfiguration->isSortedReverse()){
                array_reverse($productPositions);
            }
            $tableName = $this->resourceConnection->getTableName('catalog_category_product');
            $sql = "REPLACE $tableName (category_id, product_id, position) VALUES ";
            $this->paginateSqlInsert($sql, $categoryId, $productPositions);
            $this->updatedCategories[] = $categoryId;
        }
    }

    private function paginateSqlInsert($sql, $categoryId, $productPositions) {
        $rows = [];
        $productCount = count($productPositions);
        for ($index=1; $index <= $productCount; $index++) {
            $productId = $productPositions[$index-1];
            $rows[] = "($categoryId, $productId, $index)";
            if( $index % 100 == 0 || $index == $productCount ){
                $values = implode(', ', $rows);
                $query = $sql . $values . ';';
                $rows = [];
                $this->runSql($query);
            }
        }
    }

    private function _paginateSqlRemove($categoryId, $products){
        $perPage = 100;
        $offset = 0;
        $tableName = $this->resourceConnection->getTableName('catalog_category_product');
        $productsToDelete = array_slice($products, $offset, $perPage);
        while(count($productsToDelete)>0){
            $productsToDelete = implode(', ', $productsToDelete);
            $sql = "DELETE FROM $tableName WHERE category_id=$categoryId AND product_id IN ($productsToDelete);";
            $this->runSql($sql);
            $offset += $perPage;
            $productsToDelete = array_slice($products, $offset, $perPage);
        }
        $this->updatedCategories[] = $categoryId;
    }

    private function runSql($sql) {
        // Not for SELECT
        $conn = $this->resourceConnection->getConnection();
        $conn->query($sql);
    }
    private function runSqlSelect($sql) {
        $conn = $this->resourceConnection->getConnection();
        return $conn->fetchAll($sql);
    }
    public function reindexUpdatedCategories($categoryIds=null){
        if (isset($categoryIds)){
            array_push($this->updatedCategories, $categoryIds);
        }
        $this->updatedCategories = array_unique($this->updatedCategories);
        $this->logger->info("reindexUpdatedCategories: categoryIds: ".json_encode($this->updatedCategories));
        $indexer = $this->indexerFactory->create()->load('catalog_category_product');
        $indexer->reindexList($this->updatedCategories);
        foreach($this->updatedCategories as $categoryId) {
            // clear magento cache
            $category = $this->categoryFactory->create()->load($categoryId);
            $this->eventManager->dispatch('clean_cache_by_tags', ['object' => $category]);
        }
        $this->updatedCategories = [];
        // Tagalys custom cache clear event
        $categoryIdsObj = new \Magento\Framework\DataObject(array('category_ids' => $this->updatedCategories));
        $this->eventManager->dispatch('tagalys_category_positions_updated', ['tgls_data' => $categoryIdsObj]);
    }

    public function isTagalysCreated($category) {
        if(!is_object($category)){
            $category = $this->categoryCollection->addAttributeToSelect('parent_id')->addAttributeToFilter('entity_id', $category)->setPage(1,1)->getFirstItem();
            if(!$category->getId()){
                return false;
            }
        }
        $parentCategories = $this->getTagalysParentCategory();
        if(in_array($category->getId(), $parentCategories)){
            return true;
        }
        if($category->getId()){
            $parentId = $category->getParentId();
            if(in_array($parentId, $parentCategories)){
                return true;
            }
        }
    }
    
    public function getTagalysCreatedCategories() {
        $tagalysCreated = $this->getTagalysParentCategory();
        $tagalysLegacyCategories = $this->tagalysConfiguration->getConfig('legacy_mpage_categories', true);
        $tagalysCreated = array_merge($tagalysCreated, $tagalysLegacyCategories);
        $categories = $this->categoryCollectionFactory->create()->addAttributeToSelect('entity_id');
        foreach($categories as $category){
            if(in_array($category->getParentId(), $tagalysCreated)){
                $tagalysCreated[] = $category->getId();
            }
        }
        return $tagalysCreated;
    }

    public function getTagalysParentCategory($storeId=null){
        if (isset($storeId)){
            $categoryId = $this->tagalysConfiguration->getConfig("tagalys_parent_category_store_$storeId");
            $categoryId = $this->categoryFactory->create()->load($categoryId);
        } else {
            $categoryId = [];
            $websiteStores = $this->tagalysConfiguration->getAllWebsiteStores();
            foreach($websiteStores as $sid){
                $sid = $sid['value'];
                $catId = $this->tagalysConfiguration->getConfig("tagalys_parent_category_store_$sid");
                if($catId != null){
                    $categoryId[] = $catId;
                }
            }
        }
        return $categoryId;
    }

    public function redirectToCategoryUrl($storeId, $categoryId, $path){
        $category = $this->categoryFactory->create()->load($categoryId);
        $urlPath = $category->getUrlPath();
        if( substr($category->getUrl(), -5) === '.html' ){
            $urlPath.='.html';
        }
        $this->createUrlRewrite($storeId, $path, $urlPath, 301);
    }

    private function createUrlRewrite($storeId, $requestPath, $targetPath, $redirectType) {
        $urlRewriteModel = $this->urlRewriteFactory->create();
        $urlRewriteModel->setStoreId($storeId);
        $urlRewriteModel->setIsSystem(0);
        $urlRewriteModel->setRequestPath($requestPath);
        $urlRewriteModel->setTargetPath($targetPath);
        $urlRewriteModel->setRedirectType($redirectType);
        $urlRewriteModel->setDescription("Created by Tagalys");
        $urlRewriteModel->save();
    }

    public function updateCategoryUrlRewrite($storeId, $categoryId, $path){
        $targetPath = "catalog/category/view/id/$categoryId";
        $urlRewrite = $this->urlRewriteCollection->addStoreFilter($storeId)->addFieldToFilter('target_path', $targetPath)->addFieldToFilter('is_autogenerated', 1)->getFirstItem();
        $urlRewrite->setRequestPath($path);
        $urlRewrite->save();
    }

    private function isLegacyMpageCategory($category){
        $parentId = $category->getParentId();
        $legacyCategories = $this->tagalysConfiguration->getConfig('legacy_mpage_categories', true);
        return in_array($parentId, $legacyCategories);
    }

    public function reindexFlatCategories() {
        try{
            $categoryFlatIndexer = $this->indexerFactory->create()->load('catalog_category_flat');
            $categoryFlatIndexer->reindexAll();
        } catch(\Exception $e) {
            $this->logger->err("reindexFlatCategories: {$e->getMessage()}");
        }
    }

}
